"""Constants to be used by multiple modules of the Pantos client
library.

"""
import typing

from pantos.common.types import TokenSymbol

TOKEN_SYMBOL_PAN: typing.Final[TokenSymbol] = TokenSymbol('pan')
