"""Package for all business logic.

"""
__all__ = ['Interactor', 'InteractorError']

from pantos.client.library.business.base import Interactor
from pantos.client.library.business.base import InteractorError
